/****************************************************************************
*          J O Y S T I C K   C O N T R O L   P A N E L   A P P L E T        *  
*---------------------------------------------------------------------------*
* Written by:   Gary T. Desrosiers and Kenneth W. Clapp                     *
* Date:         May 12th, 1996                                              *
* Copyright (c) Gary T. Desrosiers, 1996. All rights reserved               *
* Copyright (c) Kenneth W. Clapp, 1996. All rights reserved                 *
*  																			*
* Description:  Implements the calibration logic for the joystick.          *
*               Interfaces with the joyio kernel driver to read             *
*               the joystick port and determine the base and delta          *
*               values used my the win32 subsystem driver to convert        *
*               the physical values returned by the joystick into           *
*               logical values. All configuration information is            *
*               written to the registry in HKEY_CURRENT_USER\Software\      *
*               Joystick.										            *
*																			*
****************************************************************************/

#include <windows.h>
#include <cpl.h>
#include "Joystick.h"
#include "resource.h"
#include <mmsystem.h>

BOOL APIENTRY JoyDlgProcTest (HWND hDlg, UINT message, UINT wParam, LONG lParam);
#define JDD_CONFIGCHANGED       0x0903

typedef struct tagApplets
{
    int icon;           // icon resource identifier
    int namestring;     // name-string resource identifier
    int descstring;     // description-string resource identifier
    int dlgtemplate;    // dialog box template resource identifier
    DLGPROC dlgfn;      // dialog box procedure
} APPLETS;

APPLETS JoyApplets[] =
{
    JOY_ICON, JOY_NAME, JOY_DESC, JOY_DLG, JoyDlgProc,

};

#define NUM_APPLETS (sizeof(JoyApplets)/sizeof(JoyApplets[0]))

UINT DevMinX[2],DevMaxX[2],DevMinY[2],DevMaxY[2];
UINT wXbase[2], wXdelta[2], wYbase[2], wYdelta[2];
UINT LogicalMin[2], LogicalMax[2];
UINT JoystickID;

JOYINFO ji;
char tempvalue[32];

HANDLE hModule = NULL;

char szCtlPanel[30];

/****************************************************************************
*
*    FUNCTION: DllMain(PVOID, ULONG, PCONTEXT)
*
*    PURPOSE: Win 32 Initialization DLL
*
*    COMMENTS:
*
*
****************************************************************************/

BOOL WINAPI DllMain(
IN PVOID hmod,
IN ULONG ulReason,
IN PCONTEXT pctx OPTIONAL)
{
    if (ulReason != DLL_PROCESS_ATTACH)
    {
        return TRUE;
    }
    else
    {
        hModule = hmod;
    }

    return TRUE;

    UNREFERENCED_PARAMETER(pctx);
}


/****************************************************************************
*
*    FUNCTION: InitJoyApplet(HWND)
*
*    PURPOSE: loads the caption string for the Control Panel
*
*    COMMENTS:
*
*
****************************************************************************/

BOOL InitJoyApplet (HWND hwndParent)
{
    LoadString (hModule, CPCAPTION, szCtlPanel, sizeof(szCtlPanel));

    return TRUE;

    UNREFERENCED_PARAMETER(hwndParent);
}


/****************************************************************************
*
*    FUNCTION: TermJoyApplet()
*
*    PURPOSE: termination procedure for the Joy applets
*
*    COMMENTS:
*
*
****************************************************************************/

void TermJoyApplet()
{
    return;
}



/****************************************************************************
*
*    FUNCTION: CPIApplet(HWND, UINT, LONG, LONG)
*
*    PURPOSE: Processes messages for control panel applets
*
*    COMMENTS:
*
*
****************************************************************************/
LONG CALLBACK CPlApplet (hwndCPL, uMsg, lParam1, lParam2)
HWND hwndCPL;       // handle of Control Panel window
UINT uMsg;          // message
LONG lParam1;       // first message parameter
LONG lParam2;       // second message parameter
{
    int iApplet;
    LPNEWCPLINFO lpNewCPlInfo;
    static iInitCount = 0;
            
    switch (uMsg) {
        case CPL_INIT:              // first message, sent once
            if (!iInitCount)
            {
                if (!InitJoyApplet(hwndCPL))
                    return FALSE;
            }
            iInitCount++;
            return TRUE;

        case CPL_GETCOUNT:          // second message, sent once
            return (LONG)NUM_APPLETS;
            break;

        case CPL_NEWINQUIRE:        // third message, sent once per app
            lpNewCPlInfo = (LPNEWCPLINFO) lParam2;

            iApplet = (int)(LONG)lParam1;
            lpNewCPlInfo->dwSize = (DWORD) sizeof(NEWCPLINFO);
            lpNewCPlInfo->dwFlags = 0;
            lpNewCPlInfo->dwHelpContext = 0;
            lpNewCPlInfo->lData = 0;
            lpNewCPlInfo->hIcon = LoadIcon (hModule,
                (LPCTSTR) MAKEINTRESOURCE(JoyApplets[iApplet].icon));
            lpNewCPlInfo->szHelpFile[0] = '\0';

            LoadString (hModule, JoyApplets[iApplet].namestring,
                        lpNewCPlInfo->szName, 32);

            LoadString (hModule, JoyApplets[iApplet].descstring,
                        lpNewCPlInfo->szInfo, 64);
            break;

        case CPL_SELECT:            // application icon selected
            break;


        case CPL_DBLCLK:            // application icon double-clicked
            iApplet = (int)(LONG)lParam1;

            DialogBox (hModule,
                       MAKEINTRESOURCE(JoyApplets[iApplet].dlgtemplate),
                       hwndCPL,
                       JoyApplets[iApplet].dlgfn);

            break;

         case CPL_STOP:              // sent once per app. before CPL_EXIT
            break;

         case CPL_EXIT:              // sent once before FreeLibrary called
            iInitCount--;
            if (!iInitCount)
                TermJoyApplet();
            break;

         default:
            break;
    }
    return 0;
}

/****************************************************************************
*
*    FUNCTION: JoyDlgProc
*
*    PURPOSE: Processes messages sent to the Joy applet.
*
*    COMMENTS:
*
*    This dialog simply puts up a box and has an OK key.  It doesn't
*    do anything except display.
*
****************************************************************************/
BOOL APIENTRY JoyDlgProc (HWND hDlg, UINT message, UINT wParam, LONG lParam)
{
static HANDLE hDriver;
static JOYINFO ji;

    switch (message)
    {
	    case WM_INITDIALOG: {
			HKEY hKey;
			DWORD size,type;
			JoystickID = JOYSTICKID1;
			DevMinX[0] = DevMinX[1] = DevMinY[0] = DevMinY[1] = 0xffffffff;
			DevMaxX[0] = DevMaxX[1] = DevMaxY[0] = DevMaxY[1] = 0;
			if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\Joystick",0,"",REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&hKey,&size) == ERROR_SUCCESS) {
				size = 8;
				if(RegQueryValueEx(hKey,"LogicalMin",0,&type,(LPBYTE)&LogicalMin,&size) != ERROR_SUCCESS) 
					LogicalMin[JOYSTICKID1] = LogicalMin[JOYSTICKID2] = 0;
				size = 8;
				if(RegQueryValueEx(hKey,"LogicalMax",0,&type,(LPBYTE)&LogicalMax,&size) != ERROR_SUCCESS)
					LogicalMax[JOYSTICKID1] = LogicalMax[JOYSTICKID2] = 65535;
				size = 8;
				if(RegQueryValueEx(hKey,"wXmin",0,&type,(LPBYTE)&DevMinX,&size) == ERROR_SUCCESS) {
					wsprintf(tempvalue,"%d",DevMinX[JoystickID]);
					SetDlgItemText(hDlg,IDE_XMIN,tempvalue);
				}
				size = 8;
				if(RegQueryValueEx(hKey,"wXmax",0,&type,(LPBYTE)&DevMaxX,&size) == ERROR_SUCCESS)  {
					wsprintf(tempvalue,"%d",DevMaxX[JoystickID]);
					SetDlgItemText(hDlg,IDE_XMAX,tempvalue);
				}
				size = 8;
				if(RegQueryValueEx(hKey,"wYmin",0,&type,(LPBYTE)&DevMinY,&size) == ERROR_SUCCESS)  {
					wsprintf(tempvalue,"%d",DevMinY[JoystickID]);
					SetDlgItemText(hDlg,IDE_YMIN,tempvalue);
				}
				size = 8;
				if(RegQueryValueEx(hKey,"wYmax",0,&type,(LPBYTE)&DevMaxY,&size) == ERROR_SUCCESS)  {
					wsprintf(tempvalue,"%d",DevMaxY);
					SetDlgItemText(hDlg,IDE_YMAX,tempvalue);
				}
				RegCloseKey(hKey);
			}
			CheckRadioButton(hDlg,IDR_JOYSTICKID1,IDR_JOYSTICKID2,IDR_JOYSTICKID1);
			wsprintf(tempvalue,"%d",LogicalMin[JoystickID]);
			SetDlgItemText(hDlg,IDE_LOGICALMIN,tempvalue);
			wsprintf(tempvalue,"%d",LogicalMax[JoystickID]);
			SetDlgItemText(hDlg,IDE_LOGICALMAX,tempvalue);
			hDriver = CreateFile("\\\\.\\joyio",GENERIC_READ|GENERIC_WRITE,FILE_SHARE_READ,NULL,OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
			SetTimer(hDlg,1,250,NULL);
			}
			return (FALSE);

		case WM_RBUTTONDBLCLK:
			MessageBox(hDlg,"Joystick/NT Driver written by Gary T. Desrosiers and Kenneth W. Clapp\n(c) Copyrighted, 1996. All rights reserved\n","Authors",MB_OK);
			break;

		case WM_TIMER:
			if(wParam == 1) {
				int BytesRead;
			if(hDriver)
				ReadFile((HANDLE)hDriver,(LPVOID)&ji,JoystickID+1,&BytesRead,NULL);
				if(ji.wXpos < DevMinX[JoystickID])
					DevMinX[JoystickID] = ji.wXpos;
				if(ji.wXpos > DevMaxX[JoystickID])
					DevMaxX[JoystickID] = ji.wXpos;
				if(ji.wYpos < DevMinY[JoystickID])
					DevMinY[JoystickID] = ji.wYpos;
				if(ji.wYpos > DevMaxY[JoystickID])
					DevMaxY[JoystickID] = ji.wYpos;
				wsprintf(tempvalue,"%d",DevMinX[JoystickID]);
				SetDlgItemText(hDlg,IDE_XMIN,tempvalue);
				wsprintf(tempvalue,"%d",DevMaxX[JoystickID]);
				SetDlgItemText(hDlg,IDE_XMAX,tempvalue);
				wsprintf(tempvalue,"%d",DevMinY[JoystickID]);
				SetDlgItemText(hDlg,IDE_YMIN,tempvalue);
				wsprintf(tempvalue,"%d",DevMaxY[JoystickID]);
				SetDlgItemText(hDlg,IDE_YMAX,tempvalue);
				InvalidateRect(hDlg,NULL,FALSE);
			}
			break;

		case WM_PAINT: {
			PAINTSTRUCT ps;
			POINT pt;
			HPEN hPen;
			RECT rect;
			HBRUSH hBrush;
			double xfactor, yfactor;
			HWND hwnd = GetDlgItem(hDlg,IDS_JOY);
			BeginPaint(hwnd,&ps);
			GetClientRect(hwnd,&rect);
			if((ji.wXpos == 5000) && (ji.wYpos == 5000)) {
				SIZE sz;
				GetTextExtentPoint32(ps.hdc,"UNPLUGGED",9,&sz);
				SetBkColor(ps.hdc,RGB(0,0,0));
				SetTextColor(ps.hdc,RGB(0,255,0));
				TextOut(ps.hdc,(rect.right-rect.left)/2-(sz.cx/2),(rect.bottom-rect.top)/2-(sz.cy/2),"UNPLUGGED",9);
				DevMinX[JoystickID] = DevMinY[JoystickID] = 0xffffffff;
				DevMaxX[JoystickID] = DevMaxY[JoystickID] = 0;
			}
			else {
				xfactor = rect.right-rect.left;
				xfactor /=(DevMaxX[JoystickID]-DevMinX[JoystickID]);
				yfactor = rect.bottom-rect.top;
				yfactor /=(DevMaxY[JoystickID]-DevMinY[JoystickID]);
				FillRect(ps.hdc,&rect,GetStockObject(BLACK_BRUSH));
				hPen = CreatePen(PS_SOLID,1,RGB(0,255,0));
				SelectObject(ps.hdc,hPen);
				wsprintf(tempvalue,"%d:%d",ji.wXpos,ji.wYpos);
				SetBkColor(ps.hdc,RGB(0,0,0));
				SetTextColor(ps.hdc,RGB(0,255,0));
				TextOut(ps.hdc,0,0,tempvalue,strlen(tempvalue));
				pt.x = (int)(xfactor*ji.wXpos);
				pt.y = (int)(yfactor*ji.wYpos);
				if((pt.x <= rect.right) && (pt.y <= rect.bottom)) {
					MoveToEx(ps.hdc,pt.x,0,NULL);
					LineTo(ps.hdc,pt.x,rect.bottom);
					MoveToEx(ps.hdc,0,pt.y,NULL);
					LineTo(ps.hdc,rect.right,pt.y);
				}
				DeleteObject(hPen);
			}
			EndPaint(hwnd,&ps);
			hBrush = CreateSolidBrush(RGB(255,0,0));
			hwnd = GetDlgItem(hDlg,IDS_BUTTON1);
			BeginPaint(hwnd,&ps);
			GetClientRect(hwnd,&rect);
			if(ji.wButtons & JOY_BUTTON1)
				FillRect(ps.hdc,&rect,hBrush);
			else
				FillRect(ps.hdc,&rect,GetStockObject(LTGRAY_BRUSH));
			EndPaint(hwnd,&ps);
			hwnd = GetDlgItem(hDlg,IDS_BUTTON2);
			BeginPaint(hwnd,&ps);
			GetClientRect(hwnd,&rect);
			if(ji.wButtons & JOY_BUTTON2)
				FillRect(ps.hdc,&rect,hBrush);
			else
				FillRect(ps.hdc,&rect,GetStockObject(LTGRAY_BRUSH));
			EndPaint(hwnd,&ps);
			DeleteObject(hBrush);
		}
		break;

		case WM_COMMAND:      
            switch(LOWORD(wParam))
            {
			case IDR_JOYSTICKID1:
				JoystickID = JOYSTICKID1;
				DevMinX[JOYSTICKID1] = DevMinY[JOYSTICKID1] = 0xffffffff;
				DevMaxX[JOYSTICKID1] = DevMaxY[JOYSTICKID1] = 0;
				break;

			case IDR_JOYSTICKID2:
				JoystickID = JOYSTICKID2;
				DevMinX[JOYSTICKID2] = DevMinY[JOYSTICKID2] = 0xffffffff;
				DevMaxX[JOYSTICKID2] = DevMaxY[JOYSTICKID2] = 0;
				break;


			case IDOK:
			case IDAPPLY: {
				HKEY hKey;
				DWORD disp;
				HANDLE hWin32Drvr;
				if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\Joystick",0,"",REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&hKey,&disp) == ERROR_SUCCESS) {
					if(GetDlgItemText(hDlg,IDE_LOGICALMIN,tempvalue,sizeof(tempvalue)) != 0)
						LogicalMin[JoystickID] = atoi(tempvalue);
					if(GetDlgItemText(hDlg,IDE_LOGICALMAX,tempvalue,sizeof(tempvalue)) != 0)
						LogicalMax[JoystickID] = atoi(tempvalue);
					wXdelta[JoystickID] = (LogicalMax[JoystickID]-LogicalMin[JoystickID])/(DevMaxX[JoystickID]-DevMinX[JoystickID]);
					wXbase[JoystickID] = (DevMinX[JoystickID] * wXdelta[JoystickID]) - LogicalMin[JoystickID];
					RegSetValueEx(hKey,"wXbase",0,REG_BINARY,(CONST BYTE*)&wXbase,8);
					RegSetValueEx(hKey,"wXdelta",0,REG_BINARY,(CONST BYTE*)&wXdelta,8);
					wYdelta[JoystickID] = (LogicalMax[JoystickID]-LogicalMin[JoystickID])/(DevMaxY[JoystickID]-DevMinY[JoystickID]);
					wYbase[JoystickID] = (DevMinY[JoystickID] * wYdelta[JoystickID]) - LogicalMin[JoystickID];
					RegSetValueEx(hKey,"wYbase",0,REG_BINARY,(CONST BYTE*)&wYbase,8);
					RegSetValueEx(hKey,"wYdelta",0,REG_BINARY,(CONST BYTE*)&wYdelta,8);
					RegSetValueEx(hKey,"LogicalMin",0,REG_BINARY,(CONST BYTE*)&LogicalMin,8);
					RegSetValueEx(hKey,"LogicalMax",0,REG_BINARY,(CONST BYTE*)&LogicalMax,8);
					RegSetValueEx(hKey,"wXmin",0,REG_BINARY,(CONST BYTE*)&DevMinX,8);
					RegSetValueEx(hKey,"wXmax",0,REG_BINARY,(CONST BYTE*)&DevMaxX,8);
					RegSetValueEx(hKey,"wYmin",0,REG_BINARY,(CONST BYTE*)&DevMinY,8);
					RegSetValueEx(hKey,"wYmax",0,REG_BINARY,(CONST BYTE*)&DevMaxY,8);
					RegCloseKey(hKey);
				}
				if(hWin32Drvr = OpenDriver(L"joystick",L"",0)) {
					SendDriverMessage(hWin32Drvr,JDD_CONFIGCHANGED,0,0);
					CloseDriver(hWin32Drvr,0,0);
				} else
					MessageBox(hDlg,"Could not inform driver of configuration changes","Driver error",MB_OK|MB_ICONEXCLAMATION);
				if(wParam == IDOK) {
					CloseHandle(hDriver);
					KillTimer(hDlg,1);
					EndDialog(hDlg, TRUE);
				}
				return (TRUE);
				}

			case IDCANCEL:
				CloseHandle(hDriver);
				KillTimer(hDlg,1);
				EndDialog(hDlg, TRUE);
				return (TRUE);
	        }

	}
    return (FALSE);
}


