/****************************************************************************
*       W I N 3 2   S U B S Y S T E M   J O Y S T I C K   D R I V E R       *  
*---------------------------------------------------------------------------*
* Written by:   Gary T. Desrosiers and Kenneth W. Clapp                     *
* Date:         May 12th, 1996                                              *
* Copyright (c) Gary T. Desrosiers, 1996. All rights reserved               *
* Copyright (c) Kenneth W. Clapp, 1996. All rights reserved                 *
*  																			*
* Description:  User mode joystick driver, called from WINMM to provide     *
*  				the functionallity of the joystick API. This code           *
*  				interfaces with the kernel driver to read the joystick      *
*  				values and button states. It also reads the configuration   *
*  				and calibration information setup by the control panel      *
*  				applet and implements dynamic installation and              *  
*               un-installation of both drivers.                            *
*                                                                           *
****************************************************************************/

#include <windows.h>
#include <mmsystem.h>
#include <mmreg.h>
#include <drvlib.h>
#include <registry.h>

TCHAR DriverName[] = TEXT("joystick");
int BytesRead;

UINT LogicalMin[2],LogicalMax[2];
INT wXbase[2], wXdelta[2];
INT wYbase[2], wYdelta[2];
UINT wXmax[2], wXmin[2];
UINT wYmax[2], wYmin[2];

DWORD type,size,disp;
HKEY hKey;
DWORD temp;
char szDir[OFS_MAXPATHNAME];

typedef struct {
		 HANDLE hDriver;
		 UINT JoystickID;
} INSTANCEDATA, *PINSTANCEDATA;

#if 0
LONG APIENTRY DefDriverProc(DWORD  dwDriverIdentifier,
                             HANDLE hDriver,
                             UINT   message,
                             LONG   lParam1,
                             LONG   lParam2);
#endif
void DbgOut(LPSTR lpszFormat, ...);

LRESULT DriverProc(DWORD dwDriverID, HDRVR hDriver, UINT uiMessage, LPARAM lParam1, LPARAM lParam2) {
   
    switch (uiMessage) {
		case JDD_CONFIGCHANGED:
			DbgOut("JDD_CONFIGCHANGED DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
			if(RegCreateKeyEx(HKEY_CURRENT_USER,TEXT("Software\\Joystick"),0,TEXT(""),REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&hKey,&disp) == ERROR_SUCCESS) {
				size = 8;RegQueryValueEx(hKey,TEXT("LogicalMin"),0,&type,(LPBYTE)&LogicalMin,&size);
				size = 8;RegQueryValueEx(hKey,TEXT("LogicalMax"),0,&type,(LPBYTE)&LogicalMax,&size);
				size = 8;RegQueryValueEx(hKey,TEXT("wXmin"),0,&type,(LPBYTE)&wXmin,&size);
				size = 8;RegQueryValueEx(hKey,TEXT("wXmax"),0,&type,(LPBYTE)&wXmax,&size);
				size = 8;RegQueryValueEx(hKey,TEXT("wYmin"),0,&type,(LPBYTE)&wYmin,&size);
				size = 8;RegQueryValueEx(hKey,TEXT("wYmax"),0,&type,(LPBYTE)&wYmax,&size);
				size = 8;RegQueryValueEx(hKey,TEXT("wXbase"),0,&type,(LPBYTE)&wXbase,&size);
				size = 8;RegQueryValueEx(hKey,TEXT("wXdelta"),0,&type,(LPBYTE)&wXdelta,&size);
				size = 8;RegQueryValueEx(hKey,TEXT("wYbase"),0,&type,(LPBYTE)&wYbase,&size);
				size = 8;RegQueryValueEx(hKey,TEXT("wYdelta"),0,&type,(LPBYTE)&wYdelta,&size);
				size = 8;RegCloseKey(hKey);
				return (LRESULT)1L;
			} else return (LRESULT)0L;

		case JDD_GETNUMDEVS:
			/*Parameters;
				LPARAM lParam1 - Not used.
				LPARAM lParam2 - Not used.
				Return Value - JDD_GETNUMDEVS returns the maximum
					number of devices supported by the driver. 
					This value may not match the number of physical 
					devices actually connected to the computer. */
			DbgOut("JDD_GETNUMDEVS DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
		    return (LRESULT)2L;

		case JDD_GETDEVCAPS: {
			UINT JoystickID;
			/*Parameters;
				LPARAM lParam1 - Specifies a pointer to a JOYCAPS 
					structure. The driver fills this structure with 
					the capabilities of the device.
				LPARAM lParam2 - Specifies the size of the structure 
					pointed to by lParam1 in bytes.
				Return Value - If JDD_GETDEVCAPS succeeds, it returns 
					MMSYSERR_NOERROR. Otherwise, it returns	JOYERR_PARMS. */
			DbgOut("JDD_GETDEVCAPS DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
			if(dwDriverID) {
				JoystickID = ((PINSTANCEDATA)dwDriverID)->JoystickID;
				((JOYCAPS*)lParam1)->wMid = MM_MICROSOFT;
				((JOYCAPS*)lParam1)->wPid = MM_PC_JOYSTICK;
				*(((JOYCAPS*)lParam1)->szPname) = 0;
				((JOYCAPS*)lParam1)->wXmin = wXmin[JoystickID];
				((JOYCAPS*)lParam1)->wXmax = wXmax[JoystickID];
				((JOYCAPS*)lParam1)->wYmin = wYmin[JoystickID];
				((JOYCAPS*)lParam1)->wYmax = wYmax[JoystickID];
				((JOYCAPS*)lParam1)->wZmin = 0;
				((JOYCAPS*)lParam1)->wZmax = 0;
				((JOYCAPS*)lParam1)->wNumButtons = 2;
				((JOYCAPS*)lParam1)->wXmin = wXmin[JoystickID];
				((JOYCAPS*)lParam1)->wPeriodMin = 0;
				((JOYCAPS*)lParam1)->wPeriodMax = 0;
			}
		    return (LRESULT)MMSYSERR_NOERROR;
			}

		case JDD_GETPOS: {
			UINT JoystickID;
			/*Parameters;
				LPARAM lParam1 - Specifies a pointer to a JOYINFO 
					structure.
				LPARAM lParam2 - Not used.
				Return Value - If the driver successfully retrieves 
					the information, JDD_GETPOS returns
					MMSYSERR_NOERROR. If the joystick is unplugged,
					JDD-GETPOS returns JOYERR_UNPLUGGED. */
			/* DbgOut("JDD_GETPOS DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);*/
			if(dwDriverID) {
				JoystickID = ((PINSTANCEDATA)dwDriverID)->JoystickID;
				ReadFile(((PINSTANCEDATA)dwDriverID)->hDriver,(LPVOID)lParam1,JoystickID+1,&BytesRead,NULL);
				if( (((JOYINFO*)lParam1)->wXpos == 5000) || (((JOYINFO*)lParam1)->wXpos == 5000))
					return (LRESULT)JOYERR_UNPLUGGED;
				((JOYINFO*)lParam1)->wXpos = ((JOYINFO*)lParam1)->wXpos*wXdelta[JoystickID] - wXbase[JoystickID];
				((JOYINFO*)lParam1)->wYpos = ((JOYINFO*)lParam1)->wYpos*wYdelta[JoystickID] - wYbase[JoystickID];
			}
		    return (LRESULT)MMSYSERR_NOERROR;
			}

		case JDD_SETCALIBRATION: {
			UINT JoystickID;
			/*Parameters
				LPARAM lParam1 - Specifies a pointer to a JOYCALIBRATE 
					structure containing the new calibration values to adopt.
				LPARAM lParam2 - Specifies a pointer to a JOYCALIBRATE 
					structure. The driver should fill this structure with 
					the existing calibration values.
				Return Value - JDD_SETCALIBRATION returns MMSYSERR_NOERROR. */
			DbgOut("JDD_SETCALIBRATION DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
			if(dwDriverID) {
				JoystickID = ((PINSTANCEDATA)dwDriverID)->JoystickID;
				((JOYCALIBRATE*)lParam2)->wXbase = wXbase[JoystickID];
				((JOYCALIBRATE*)lParam2)->wXdelta = wXdelta[JoystickID];
				((JOYCALIBRATE*)lParam2)->wYbase = wYbase[JoystickID];
				((JOYCALIBRATE*)lParam2)->wYdelta = wYdelta[JoystickID];
				((JOYCALIBRATE*)lParam2)->wZbase = 0;
				((JOYCALIBRATE*)lParam2)->wZdelta = 0;
				wXbase[JoystickID] = ((JOYCALIBRATE*)lParam1)->wXbase;
				wXdelta[JoystickID] = ((JOYCALIBRATE*)lParam1)->wXdelta;
				wYbase[JoystickID] = ((JOYCALIBRATE*)lParam1)->wYbase;
				wYdelta[JoystickID] = ((JOYCALIBRATE*)lParam1)->wYdelta;
			}
		    return (LRESULT)MMSYSERR_NOERROR;
			}

        case DRV_LOAD:
			DbgOut("DRV_LOAD DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
			return (LRESULT) SendDriverMessage(hDriver,JDD_CONFIGCHANGED,0,0);

        case DRV_FREE:
			DbgOut("DRV_FREE DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
            return (LRESULT)1L;

        case DRV_OPEN: {
			PINSTANCEDATA pInst = (PINSTANCEDATA)GlobalAlloc(GMEM_FIXED,sizeof(INSTANCEDATA));
			DbgOut("DRV_OPEN DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
			pInst->hDriver = CreateFile(TEXT("\\\\.\\joyio"),GENERIC_READ|GENERIC_WRITE,FILE_SHARE_READ,NULL,OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
			pInst->JoystickID = lParam2;
            return (LRESULT)pInst;
		    }

        case DRV_CLOSE: {
			PINSTANCEDATA pInst = (PINSTANCEDATA)dwDriverID;
			DbgOut("DRV_CLOSE DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
			if(pInst->hDriver)
				CloseHandle((HANDLE)pInst->hDriver);
			GlobalFree(pInst);
            return (LRESULT)1L;
		    }

        case DRV_ENABLE:
			DbgOut("DRV_ENABLE DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
            return 1L;

        case DRV_DISABLE:
			DbgOut("DRV_DISABLE DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
            return (LRESULT)1L;

        case DRV_QUERYCONFIGURE:
			DbgOut("DRV_QUERYCONFIGURE DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
            return 0L; /* Configure not supported */

        case DRV_CONFIGURE:
			DbgOut("DRV_CONFIGURE DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
            return 1L;

        case DRV_INSTALL:
			DbgOut("DRV_INSTALL DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
			if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,TEXT("SYSTEM\\CurrentControlSet\\Services\\joyio"),0,TEXT(""),REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&hKey,&disp) == ERROR_SUCCESS) {
				temp = SERVICE_KERNEL_DRIVER;
				RegSetValueEx(hKey,TEXT("Type"),0,REG_DWORD,(CONST BYTE*)&temp,4);
				temp = SERVICE_AUTO_START;
				RegSetValueEx(hKey,TEXT("Start"),0,REG_DWORD,(CONST BYTE*)&temp,4);
				temp = SERVICE_ERROR_NORMAL;
				RegSetValueEx(hKey,TEXT("ErrorControl"),0,REG_DWORD,(CONST BYTE*)&temp,4);
				RegSetValueEx(hKey,TEXT("Group"),0,REG_SZ,(CONST BYTE*)TEXT("Extended Base"),26);
				RegCloseKey(hKey);
			}
			GetWindowsDirectoryA(szDir,OFS_MAXPATHNAME);
			strcat(szDir,"\\system.ini");
			WritePrivateProfileStringA("drivers","joystick","joystick.dll",szDir);
            return (LRESULT)DRV_RESTART;

        case DRV_REMOVE:
			DbgOut("DRV_REMOVE DrvrID=%X hDrvr=%X Msg=%X lparam1=%X lParam2=%X",dwDriverID,hDriver,uiMessage,lParam1,lParam2);
			if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,TEXT("SYSTEM\\CurrentControlSet\\Services"),0,TEXT(""),REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&hKey,&disp) == ERROR_SUCCESS) {
				RegDeleteKey(hKey,TEXT("joyio"));
			    RegCloseKey(hKey);
			}
			if(RegCreateKeyEx(HKEY_CURRENT_USER,TEXT("Software"),0,TEXT(""),REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&hKey,&disp) == ERROR_SUCCESS) {
				RegDeleteKey(hKey,TEXT("Joystick"));
				RegCloseKey(hKey);
			}
            return 1L;

        default:
            return DefDriverProc(dwDriverID, hDriver,uiMessage,lParam1,lParam2);
    }
}

void DbgOut(LPSTR lpszFormat, ...)
{
    char buf[256];
    va_list va;

    OutputDebugStringA("JOYSTICK: ");

    va_start(va, lpszFormat);
    vsprintf(buf, lpszFormat, va);
    va_end(va);

    OutputDebugStringA(buf);
    OutputDebugStringA("\n");
}
